unit USunriseChatCore;

interface

uses
  SysUtils, Classes, ExtCtrls, cWindows, cSockets, cSocketsUDP, cTCPClient,
  cTCPServer, Forms, Dialogs, cWinSock, cSocketHostLookup, WinSock, Graphics,
  Windows, Registry, cUtils, typinfo;

const
  ProtocolVersion = 2;
  CoreVersion = '2.7';
  RemoveTimeout = 33/24/3600; // 33 seconds

  clSysMessage = $7fffffff;
  riAll = -1;

  // User status modes
  usAway = 'Away';
  usOnline = 'Online';
  usWritting = 'Writting';
  usInvisible = 'Invisible';
  usOffline = 'Offline';

type
  TAppEventType = (aeCommonMessage, aeOneUserMessage, aeUserConnect,
    aeUserDisconnect, aeUserRestore, aeUserTimeout, aeUserGoAway, aeUserGoBack,
    aeSendCallUp, aeReceiveCallUp, aeSendPing, aePingResponse, aeUserChangeNick,
    aeUserChangeNickConflict, aeCreateRoom, aeLeaveRoom, aeUserAutoGoAway,
    aeUserAutoGoBack, aeShowHours, aeStart);

  TMessageEvent = procedure (Text: string; Color: Integer) of object;
  TClassMethod = procedure of object;
  TArrayOfString = array of string;

  TNetworkInterface = record
    IPAddress: string;
    BroadcastIPAddress: string;
    SubNetMask: string;
    Name: string;
    DeviceName: string;
    GUID: string;
  end;

  TRoomType = (rtPublic, rtPrivate);
  TLine = record
    Text: string;
    Font: TFont;
    Time: TDateTime;
    Nick: string;
    EventType: TAppEventType;
  end;

  TRoom = record
    Name: string;
    Typ: TRoomType;
    StartLine: Integer;
    Count: Integer;
    Lines: array of TLine;
  end;

  TUser = record
    Nick: string;
    HostName: string;
    IP: string;
    Female: Boolean;
    Color: Integer;
    Client: string;
    ClientVersion: string;
    CoreVersion: string;
    OSVersion: string;
    OSUser: string;
    Status: string;
    LastTime: TDateTime;
    UpTime: TDateTime;
    LocalSystemTime: TDateTime;
    UserInfoTime: TDateTime;
    ID: Integer;
    NickTime: TDateTime;
    Reason: string;
    Sequence: Integer;
    ErrorCount: Integer;
    BlockMessages: Boolean;
    Delay: TDateTime;
    RoomName: string;
    RoomType: TRoomType;
    DetailInfo: string;
    IdleTime: Integer;
  end;

  TAddMessageMethod = procedure(RoomIndex: Integer; const Args: array of const; EventType: TAppEventType; TextColor: TColor = clSysMessage; Nick: string = '') of object;

  TSunriseChatCore = class(TDataModule)
    fndTCPServer1: TfndTCPServer;
    fndTCPClient1: TfndTCPClient;
    fndUDPClientSocket1: TfndUDPClientSocket;
    Timer1: TTimer;
    Timer2: TTimer;
    Timer3: TTimer;
    procedure fndTCPServer1DataAvailable(Sender: TTCPServerClient);
    procedure fndTCPClient1ConnectFailed(Sender: ATCPClient);
    procedure fndTCPClient1DataAvailable(Sender: ATCPClient);
    procedure fndUDPClientSocket1DataAvailable(Sender: AUDPSocket);
    procedure fndTCPClient1Close(Sender: ATCPClient);
    procedure Timer1Timer(Sender: TObject);
    procedure Timer2Timer(Sender: TObject);
    procedure DataModuleCreate(Sender: TObject);
    procedure fndUDPClientSocket1Error(Sender: ASocket);
    procedure Timer3Timer(Sender: TObject);
  private
    FOnAddMessage: TAddMessageMethod;
    FOnChangeNetworkState: TClassMethod;
    FOnUserListChange: TClassMethod;
//    FOnPingResponse: TClassMethod;
//    FOnCallUp: TClassMethod;
    FOnCommand: TClassMethod;
//    FOnNickChange: TClassMethod;
    FConnected: Boolean;
    FOnRoomListChanged: TClassMethod;
    FMaxRoomLines: Integer;
    LibHandle: HInst;   //dll handle
    LastHour: Word;
    procedure ExecuteCommand(Text: string);
    function GetWindowsVersionStr: string;
    procedure SetMaxRoomLines(Number: Integer);
    function SecondsIdle: DWord;
    function Explode(Separator: Char; Data: string): TArrayOfString;
  public
    NetworkInterfaces: array of TNetworkInterface;
    UDPPort: Integer;
    NetworkInterfaceIPAddress: string;
    LocalUser: TUser;
    UserList: array of TUser;
    ActiveRoomIndex: Integer;
    RoomList: array of TRoom;
    LastMessage: string;
    LastFont: TFont;
    LastUserIndex: Integer;
    LastEventType: TAppEventType;
    AutoAwayIdleTime: Integer;
    NetworkBridge: Boolean; 
    NetworkInterfaceIndex: Integer;
    procedure ChangeNick(NewNick: string; SendCmd: Boolean = True);
    procedure Init;
    procedure DeleteRoom(RoomIndex: Integer);
    procedure AddMessage(RoomIndex: Integer; const Args: array of const; EventType: TAppEventType; TextColor: TColor = clSysMessage; Nick: string = '');
    procedure DeleteOfflineUsers;
    procedure UpdateUser(User: TUser);
    constructor Create(AOwner: TComponent); override;
    procedure SendCommand(Command, Text: string; DestinationIP: string = ''; DestinationID: Integer = 0);
    procedure StartNetwork;
    destructor Done;
    property Connected: Boolean read FConnected;
    procedure LoadNetworkInterfaceList;
    procedure SelectNetworkInterface(SelectedGUID: string);
  published
    property MaxRoomLines: Integer read FMaxRoomLines write SetMaxRoomLines;
    property LocalIPAddress: string read LocalUser.IP write LocalUser.IP;
//    property OnNickChange: TClassMethod read FOnNickChange write FOnNickChange;
    property OnUserListChange: TClassMethod read FOnUserListChange write FOnUserListChange;
    property OnAddMessage: TAddMessageMethod read FOnAddMessage write FOnAddMessage;
    property OnChangeNetworkState: TClassMethod read FOnChangeNetworkState write FOnChangeNetworkState;
//    property OnPingResponse: TClassMethod read FOnPingResponse write FOnPingResponse;
//    property OnCallUp: TClassMethod read FOnCallUp write FOnCallUp;
    property OnCommand: TClassMethod read FOnCommand write FOnCommand;
    property OnRoomListChanged: TClassMethod read FOnRoomListChanged write FOnRoomListChanged;
  end;

var
  SunriseChatCore: TSunriseChatCore;
  GetLastInputInfo2: function(var plii: TLastInputInfo): BOOL; stdcall;

implementation

uses DateUtils, UMainWindow, StrUtils;

{$R *.dfm}

{ TSunriseChatCore }

function TSunriseChatCore.SecondsIdle: DWord;
var
   liInfo: TLastInputInfo;
begin
  // Only WinNT+
 if Assigned(GetLastInputInfo2) then begin
    liInfo.cbSize := SizeOf(TLastInputInfo) ;
    GetLastInputInfo2(liInfo) ;
    Result := (GetTickCount - liInfo.dwTime) DIV 1000;
  end else Result:= 0;
//  MainWindow.RichEdit1.Lines.Add(IntToStr(Result));
//  SysUtils.Beep;
//  Result:= IdleTrackerGetLastTickCount div 1000;
//  Button1.Caption:= IntToStr(Result);
//  Result:= 0;
end;

constructor TSunriseChatCore.Create(AOwner: TComponent);
begin
  inherited;

  // Try to load indirectly function GetLastInputInfo for compatibility with Win9x
  LibHandle:= LoadLibrary('user32.dll');
  if LibHandle = 0 then begin
    GetLastInputInfo2:= nil;
  end else begin
    @GetLastInputInfo2:= GetProcAddress(LibHandle,'GetLastInputInfo');
    if (@GetLastInputInfo2 = nil) then FreeLibrary(LibHandle);
  end;

  with LocalUser do begin
    OSVersion:= GetWindowsVersionStr;
    OSUser:= cWindows.GetUserName;
    Status:= usOnline;
    HostName:= LocalHostName;
    Uptime:= Now;
    IP:= LocalIPAddress;
    ID:= Application.Handle;
    RoomName:= 'Vichni';
    RoomType:= rtPublic;
  end;
  LocalUser.CoreVersion:= CoreVersion;
  SetLength(RoomList,1);
  with RoomList[0] do begin
    Name:= 'Vichni';
    Typ:= rtPublic;
//    AddMessage(0,'Vtejte v SunriseChatu!');
  end;
(*
  with RoomList[1] do begin
    Name:= 'George';
    Typ:= rtPrivate;
    ActiveRoomIndex:= 1;
    //AddMessage(1,'Test');
  end;
*)
end;

destructor TSunriseChatCore.Done;
begin

end;

procedure TSunriseChatCore.ExecuteCommand(Text: string);
var
  RemoteUser: TUser;
  ProtocolVersion: Integer;
  DestinationIP: string;
  DestinationID: Integer;
  Seq: Integer;
  Command: string;
  Data: string;
  NewNick: string;
  I: Integer;
  IdleTime2: Integer;

function Parse: string;
begin
  Result:= Copy(Text,1,Pos('|',Text)-1);
  Delete(Text,1,Length(Result)+1);
end;

begin
//  OnAddMessage(Text,clSysMessage);
  //AddMessage(Command);
  with SunriseChatCore, RemoteUser do try
    ProtocolVersion:= StrToInt(Parse);
    case ProtocolVersion of
      1: begin
        Client:= Parse;
        ClientVersion:= Parse;
        Nick:= Parse;
        OSVersion:= Parse;
        ID:= StrToInt(Parse);
        Color:= StrToInt(Parse);
        Uptime:= StrToDateTime(Parse);
        IP:= Parse;
        HostName:= Parse;
        Command:= Parse;
        Data:= Parse;
        if Command = 'Message' then begin
//          if not BlockMessages then OnAddMessage('(Nepodporovan verze) <'+Nick+'> '+Data,Color);
        end else
      end;
      2: begin
        DestinationIP:= Parse;            // Destination IP
        DestinationID:= StrToInt(Parse);            // Destination ID
        IP:= Parse;                       // Source IP
        ID:= StrToInt(Parse);             // Source ID
        Seq:= StrToInt(Parse);       // sequence command number

        // Load remote user data
        I:= 0;
        while (I<Length(UserList)) and ((UserList[I].IP<>IP) or (UserList[I].ID<>ID)) do I:= I + 1;
        if I<Length(UserList) then begin
          RemoteUser:= UserList[I];
        end else Sequence:= Seq-1;
        if Seq<>(RemoteUser.Sequence+1) then ErrorCount:= ErrorCount + 1;
        Sequence:= Seq;
        UpdateUser(RemoteUser);

        if ((DestinationIP = LocalUser.IP) and (DestinationID = LocalUser.ID)) or (DestinationIP='') then begin

          Command:= Parse;                  // Command
          if Command = 'Message' then begin
            Color:= StrToInt(Parse);
            LastMessage:= Parse;
            RoomName:= Parse;
            RoomType:= TRoomType(StrToInt(Parse));

            // Search room
            I:= 0;
            while (I<Length(RoomList)) and (RoomList[I].Name<>RoomName) do I:= I + 1;
            if (RoomType = rtPublic) and (I=Length(RoomList)) then begin
              SetLength(RoomList,Length(RoomList)+1);
              with RoomList[High(RoomList)] do begin
                Name:= RoomName;
                Typ:= RoomType;
                SetLength(Lines,FMaxRoomLines);
                StartLine:= 0;
                Count:= 0;
              end;
              if Assigned(FOnRoomListChanged) then FOnRoomListChanged;
            end;

            UpdateUser(RemoteUser);
            if (I<Length(RoomList)) and not BlockMessages then begin
              AddMessage(I, [RemoteUser.Nick, LastMessage], aeCommonMessage,Color,Nick);
            end;
          end else
          if Command = 'CreateRoom' then begin
            RoomName:= Parse;
            RoomType:= TRoomType(StrToInt(Parse));
            //ShowMessage(LocalUser.IP+','+IP+' '+IntToStr(LocalUser.ID)+','+IntToStr(ID));
            if (RoomType = rtPublic) or ((RoomType = rtPrivate) and (LocalUser.IP = DestinationIP) and (LocalUser.ID = DestinationID)) then begin
              // Search room
              I:= 0;
              while (I<Length(RoomList)) and (RoomList[I].Name<>RoomName) do I:= I + 1;
              if I<Length(RoomList) then
              else begin
                SetLength(RoomList,Length(RoomList)+1);
                with RoomList[High(RoomList)] do begin
                  Name:= RoomName;
                  Typ:= RoomType;
                  SetLength(Lines,FMaxRoomLines);
                  StartLine:= 0;
                  Count:= 0;
                end;
              end;
              //ActiveRoomIndex:= I;
              if Assigned(FOnRoomListChanged) then FOnRoomListChanged;
              AddMessage(riAll, [RemoteUser.Nick, RoomName] ,aeCreateRoom);
            end;
          end else
          if Command = 'LeaveRoom' then begin
            RoomName:= Parse;

            // Search room
            I:= 0;
            while (I<Length(RoomList)) and (RoomList[I].Name<>RoomName) do I:= I + 1;
            if I<Length(RoomList) then
              AddMessage(riAll, [RemoteUser.Nick, RoomName], aeLeaveRoom);
          end else
          if Command = 'Connect' then begin
            Nick:= Parse;
            NickTime:= StrToDateTime(Parse);
            if (LocalUser.IP<>IP) or (LocalUser.ID<>ID) then
            AddMessage(riAll,[Nick, HostName], aeUserConnect);
            SendCommand('UserInfo','');
            UpdateUser(RemoteUser);
          end else
          if Command = 'Disconnect' then begin
            AddMessage(riAll,[Nick],aeUserDisconnect);

            // Delete disconnected user
            I:= 0;
            while (I<Length(UserList)) and ((UserList[I].IP <> RemoteUser.IP) or (UserList[I].ID <> RemoteUser.ID)) do I:= I + 1;
            if I < Length(UserList) then begin
              UserList[I]:= UserList[High(UserList)];
              SetLength(UserList,High(UserList));
            end;
            if Assigned(OnUserListChange) then OnUserListChange;
          end else
          if Command = 'GoAway' then begin
            Reason:= Parse;
            Status:= usAway;
            AddMessage(riAll, [Nick, Reason], aeUserGoAway);
            UpdateUser(RemoteUser);
          end else
          if Command = 'GoOnline' then begin
            Status:= usOnline;
            AddMessage(riAll, [Nick], aeUserGoBack);
            UpdateUser(RemoteUser);
          end else
          if Command = 'CallUp' then begin
            if not BlockMessages then AddMessage(riAll, [Nick], aeReceiveCallUp);
          end else
          if Command = 'Ping' then begin
            SendCommand('PingResponse',Parse,IP,ID);
          end else
          if Command = 'PingResponse' then begin
            AddMessage(riAll,[Nick,TimeToStr(Now-StrToTime(Parse))],aePingResponse);
          end else
          if Command = 'Alive' then begin
            Delay:= Now-LastTime;
            LastTime:= Now;
            if not TryStrToInt(Parse,IdleTime2) then IdleTime2:= 0;
            if (IdleTime2 > AutoAwayIdleTime*60) and (IdleTime < AutoAwayIdleTime*60) and (Status = usOnline) then begin
              AddMessage(riAll,[Nick, IntToStr(AutoAwayIdleTime)], aeUserAutoGoAway);
            end;
            if (IdleTime > AutoAwayIdleTime*60) and (IdleTime2 < AutoAwayIdleTime*60) and (Status = usOnline) then begin
              AddMessage(riAll, [Nick, IntToStr(IdleTime div 60)], aeUserAutoGoBack);
            end;
            IdleTime:= IdleTime2;
            UpdateUser(RemoteUser);
          end else
          if Command = 'WhoIs' then begin
            SendCommand('UserInfo','');
          end else
          if Command = 'UserInfo' then begin
            NewNick:= Parse;
            if (NewNick <> Nick) and (NickTime<>0) then AddMessage(riAll, [Nick, NewNick], aeUserChangeNick);
            Nick:= NewNick;
            NickTime:= StrToDateTime(Parse);
            if (HostName='') and (LastTime<>0) then AddMessage(riAll, [RemoteUser.Nick], aeUserRestore);
            Reason:= Parse;
            Status:= Parse;
            HostName:= Parse;
            OSVersion:= Parse;
            OSUser:= Parse;
            Uptime:= StrToDateTime(Parse);
            Client:= Parse;
            ClientVersion:= Parse;
            CoreVersion:= Parse;
            if Length(Text)>0 then begin
              LocalSystemTime:= StrToDateTime(Parse);
            end;
            DetailInfo:= Parse;

            UserInfoTime:= Now;

            // Nick conflict test
            if (LocalUser.Nick = Nick) and ((LocalUser.IP <> IP) or (LocalUser.ID <> ID)) then begin
              if LocalUser.NickTime > NickTime then begin
                // Set nick to Host(Guest) n
                if Copy(LocalUser.Nick,1,4) = 'Host' then begin
                  if TryStrToInt(Copy(LocalUser.Nick,5,255),I) then I:= I + 1 else I:= 1;
                end else I:= 1;
                ChangeNick('Host' + IntToStr(I), False);
                AddMessage(riAll, [LocalUser.Nick], aeUserChangeNickConflict);
              end else SendCommand('UserInfo','');
            end;
            UpdateUser(RemoteUser);
          end;
          if Nick='' then SendCommand('WhoIs','',IP,ID);
        end;
      end;
    end;
  except
  end;
end;

function UserIdleTime: Integer;
begin
end;

procedure TSunriseChatCore.SendCommand(Command, Text: string; DestinationIP: string = ''; DestinationID: Integer = 0);
var
  Data: string;

procedure AddPart(Part: string);
begin
  Data:= Data + Part + '|';
end;

begin
  if FConnected and (LocalUser.Status <> usInvisible) then begin
    //if not NoResloveHostName then
    Data:= '';
    with LocalUser do begin
      AddPart(IntToStr(ProtocolVersion));// Version of protocol
      AddPart(DestinationIP);            // Destination IP
      AddPart(IntToStr(DestinationID));  // Destination ID
      AddPart(IP);                       // Source IP
      AddPart(IntToStr(ID));             // Source ID
      Sequence:= Sequence + 1;
      AddPart(IntToStr(Sequence));       // Sequence command number
      AddPart(Command);                  // Command
      if Command = 'Message' then begin
        AddPart(IntToStr(Color));        // User text color
        AddPart(Text);                   // Command data
        AddPart(RoomList[ActiveRoomIndex].Name);               // Room name
        AddPart(IntToStr(Integer(RoomList[ActiveRoomIndex].Typ))); // Public or private room
      end else
      if Command = 'Connect' then begin
        AddPart(Nick);
        AddPart(DateTimeToStr(NickTime));
      end else
      if Command = 'Disconnect' then begin
      end else
      if Command = 'GoAway' then begin
        AddPart(Reason);                 // Away mode reason
      end else
      if Command = 'GoOnline' then begin
      end else
      if Command = 'GoAutoAway' then begin
      end else
      if Command = 'GoAutoOnline' then begin
      end else
      if Command = 'CallUp' then begin
      end else
      if Command = 'CallUpResponse' then begin
      end else
      if Command = 'Ping' then begin
        AddPart(TimeToStr(Now));         // Send time
      end else
      if Command = 'PingResponse' then begin
        AddPart(Text);                   // Send time
      end else
      if Command = 'Alive' then begin
        AddPart(IntToStr(SecondsIdle));
      end else
      if Command = 'WhoIs' then begin
      end else
      if Command = 'CreateRoom' then begin
        AddPart(RoomName);                    // Room name
        AddPart(IntToStr(Integer(RoomType))); // Public or private room
      end else
      if Command = 'LeaveRoom' then begin
        AddPart(RoomList[ActiveRoomIndex].Name);                    // Room name
      end else
      if Command = 'UserInfo' then begin
        AddPart(Nick);                   // User name
        AddPart(DateTimeToStr(NickTime));    // User name time
        AddPart(Reason);                 // Away mode reason
        AddPart(Status);                 // User status
        AddPart(HostName);               // Local host name
        AddPart(OSVersion);              // OS version
        AddPart(OSUser);                 // Logged user
        AddPart(DateTimeToStr(Uptime));  // Application uptime
        AddPart(Client);                 // Name of client application
        AddPart(ClientVersion);          // Version of application
        AddPart(CoreVersion);            // SunriseChatCoreVersion
        AddPart(DateTimeToStr(Now));     // Local system time
        AddPart(DetailInfo);             // Detailed info about user
      end;
    end;
    fndTCPClient1.Socket.SendStr(Data+#13);
  end;
end;

procedure TSunriseChatCore.fndTCPServer1DataAvailable(
  Sender: TTCPServerClient);
var
  Text: string;
begin
  if Sender.Socket.InBufferSize>0 then begin
    Text:= Sender.Socket.ReadAvailable;
    fndUDPClientSocket1.Broadcast(IntToStr(UDPPort),Text);
//    fndUDPClientSocket1.SendStr(Text);
  end;
end;

procedure TSunriseChatCore.fndTCPClient1ConnectFailed(Sender: ATCPClient);
begin
//  ShowMessage(Sender.Socket.ErrorMessage);
end;

procedure TSunriseChatCore.fndTCPClient1DataAvailable(Sender: ATCPClient);
var
  Data, Text: string;
begin
  if Sender.Socket.InBufferSize>0 then begin
    Data:= Sender.Socket.ReadAvailable;
    if Pos(#13,Data)>0 then begin
      repeat
        Text:= Copy(Data,1,Pos(#13,Data)-1);
        Delete(Data,1,Length(Text)+1);
        ExecuteCommand(Text);
      until (Data='') or (Pos(#13,Data)=0);
    end else ExecuteCommand(Data);
  end;
end;

procedure TSunriseChatCore.fndUDPClientSocket1DataAvailable(
  Sender: AUDPSocket);
var
  Text: string;
  Address : TSockAddr;
  I: Integer;
  LocalIPs: TInAddrArray;
begin
  if FConnected then begin
    Sender.ReadPacket(Text,Address);

//  ShowMessage(Text);
    for I:= 0 to fndTCPServer1.ClientCount-1 do
      with fndTCPServer1.Client[I] do Socket.SendStr(Text);
  end;
end;

procedure TSunriseChatCore.fndTCPClient1Close(Sender: ATCPClient);
begin
//  ShowMessage('TCPClient close');
  FConnected := False;
  Timer1.Enabled:= True;
  if Assigned(FOnChangeNetworkState) then FOnChangeNetworkState;
end;

procedure TSunriseChatCore.StartNetwork;
var
  I, II: Integer;
const
  Stav: array [0..6] of string = ('Closed', 'Resolving', 'Resolved', 'Connecting',
                  'Negotiating', 'Connected', 'Listening');
begin
  fndTCPClient1.OnClose:= nil;
  Timer1.Enabled:= False;
  FConnected:= False;
  fndTCPClient1.Active:= False;
  with fndTCPServer1 do begin
    Active:= False;
    ListenPort:= IntToStr(UDPPort+1);
    LocalHost:= LocalUser.IP; //'localhost';
    try
      Active:= True;
    except
    end;
  //  ShowMessage(Stav[Integer(fndTCPServer1.Socket.State)]);

    fndUDPClientSocket1.Terminate;
    fndUDPClientSocket1.LocalHost:= LocalUser.IP;
    fndUDPClientSocket1.LocalPort:= IntToStr(UDPPort);
    fndUDPClientSocket1.Host:= NetworkInterfaces[NetworkInterfaceIndex].BroadcastIPAddress;
    fndUDPClientSocket1.Port:= IntToStr(UDPPort);
    fndUDPClientSocket1.BroadcastOption:= True;
    try
      fndUDPClientSocket1.Bind;
    except
//      on EWinSock do begin
//        ShowMessage('Nelze inicializovat s!');
//        Application.Terminate;
//      end;
    end;
  end;
  with fndTCPClient1 do begin
    Active:= False;
    Host:= LocalUser.IP; // 'localhost';
    Port:= IntToStr(UDPPort+1);
    LocalHost:= LocalUser.IP; //'localhost';
    I:= 2;
    repeat
      Active:= False;
      LocalPort:= IntToStr(UDPPort+I);
      I:= I + 1;
      try
        Active:= True;
      except
      end;
      II:= 0;
      while (Socket.State in [ssResolving, ssNegotiating, ssConnecting]) and (II<100) do begin
        Application.ProcessMessages;
        //Sleep(10);
        II:= II + 1;
      end;
//     ShowMessage(Stav[Integer(fndTCPClient1.Socket.State)]);
    until (Socket.State in [ssConnected, ssResolved]) or (I>30); //or (fndTCPServer1.Socket.State = ssClosed);
    FConnected := not (fndUDPClientSocket1.Bound xor (fndTCPServer1.Socket.State = ssListening)) and
    (fndTCPClient1.Socket.State in [ssConnected, ssResolved]);
    //ShowMessage(IntToStr(Integer(fndTCPClient1.Socket.State))+','+BoolToStr(fndUDPClientSocket1.Bound)
    //+','+BoolToStr(FConnected)+','+BoolToStr(fndTCPClient1.Socket.Connected));
    if FConnected then begin
      SunriseChatCore.SendCommand('UserInfo','');
      SunriseChatCore.SendCommand('Connect','');
    end;
    if Assigned(FOnChangeNetworkState) then OnChangeNetworkState;
  end;
  Timer1.Enabled := True;
  fndTCPClient1.OnClose:= fndTCPClient1Close;
end;

procedure TSunriseChatCore.Timer1Timer(Sender: TObject);
begin
//  Timer1.Enabled := False;
//  if fndUDPClientSocket1.State
  if not FConnected then StartNetwork;
  if FConnected and (HourOf(Now) <> LastHour) then begin
    LastHour:= HourOf(Now);
    AddMessage(riAll,[HourOf(Now),MinuteOf(Now)], aeShowHours);
  end;
//  Timer1.Enabled := True;
end;

procedure TSunriseChatCore.Init;
begin
  StartNetwork;
end;

function TSunriseChatCore.GetWindowsVersionStr: string;
begin
  case Win32Platform of
    1: case Win32MajorVersion of
      4: case Win32MinorVersion of
        0: Result:= '95';
        10: Result:= '98';
        90: Result:= 'Me';
      end;
    end;
    2: case Win32MajorVersion of
      4: case Win32MinorVersion of
        0: Result:= 'NT 4.0';
      end;
      5: case Win32MinorVersion of
        0: Result:= '2000';
        1: Result:= 'XP';
        2: Result:= 'Server 2003';
      end;
    end;
  end;
  Result:= 'Windows ' + Result;
end;

procedure TSunriseChatCore.DeleteOfflineUsers;
var
  I: Integer;
  Changed: Boolean;
begin
  I:= 0;
  Changed:= False;
  while (I<Length(UserList)) do begin
    if (Now - UserList[I].LastTime)>RemoveTimeout then begin
      AddMessage(riAll, [UserList[I].Nick], aeUserTimeout);
      UserList[I]:= UserList[High(UserList)];
      SetLength(UserList,High(UserList));
      Changed:= True;
    end else I:= I + 1;
  end;
  if Changed and Assigned(FOnUserListChange) then FOnUserListChange;
//  TSocketState
end;

procedure TSunriseChatCore.UpdateUser(User: TUser);
var
  I: Integer;
begin
  I:= 0;
  while (I<Length(UserList)) and ((UserList[I].IP<>User.IP) or ((UserList[I].ID<>User.ID))) do I:= I + 1;
  if I = Length(UserList) then begin
    SetLength(UserList,Length(UserList)+1);
  end;
  User.LastTime:= Now;
  UserList[I]:= User;
  LastUserIndex:= I;
  if Assigned(FOnUserListChange) then FOnUserListChange;
end;

procedure TSunriseChatCore.Timer2Timer(Sender: TObject);
begin
//  MainWindow.Label1.Caption:= IntToStr(Integer(fndTCPClient1.Socket.State));
  SendCommand('Alive','');
  DeleteOfflineUsers;
end;

procedure TSunriseChatCore.AddMessage(RoomIndex: Integer; const Args: array of const; EventType: TAppEventType; TextColor: TColor = clSysMessage; Nick: string = '');
var
  I: Integer;
begin
  if Assigned(OnAddMessage) then OnAddMessage(RoomIndex, Args, EventType, TextColor, Nick);

  if (LastMessage<>'') and (RoomIndex < Length(RoomList)) then begin
    if RoomIndex = riAll then begin
      for RoomIndex := 0 to High(RoomList) do with RoomList[RoomIndex] do begin
        Count := Count + 1;
        if Count > FMaxRoomLines then begin
          StartLine := (StartLine + 1) mod FMaxRoomLines;
          Count := FMaxRoomLines;
        end;
        I := (StartLine + Count - 1) mod FMaxRoomLines;
        Lines[I].Text := LastMessage;
        Lines[I].Font := LastFont;
        Lines[I].Time := Now;
        Lines[I].Nick := Nick;
      end;
    end else with RoomList[RoomIndex] do begin
      Count := Count + 1;
      if Count > FMaxRoomLines then begin
        StartLine := (StartLine + 1) mod FMaxRoomLines;
        Count := FMaxRoomLines;
      end;
      I := (StartLine + Count - 1) mod FMaxRoomLines;
      Lines[I].Text := LastMessage;
      Lines[I].Font := LastFont;
      Lines[I].Time := Now;
      Lines[I].Nick := Nick;
    end;
  end;
end;

procedure TSunriseChatCore.DeleteRoom(RoomIndex: Integer);
begin
  RoomList[RoomIndex]:= RoomList[High(RoomList)];
  SetLength(RoomList,High(RoomList));
  FOnRoomListChanged;
end;

procedure TSunriseChatCore.ChangeNick(NewNick: string; SendCmd: Boolean = True);
begin
  with SunriseChatCore do begin
    LocalUser.NickTime:= Now;
    LocalUser.Nick:= NewNick;
    SendCommand('UserInfo','');
  end;
end;

procedure TSunriseChatCore.SetMaxRoomLines(Number: Integer);
var
  I: Integer;
begin
  FMaxRoomLines:= Number;
  for I:= 0 to High(RoomList) do with RoomList[I] do begin
    SetLength(Lines,Number);
    StartLine:= 0;
    Count:= 0;
  end;
end;

procedure TSunriseChatCore.DataModuleCreate(Sender: TObject);
begin
  LastHour:= HourOf(Now);
  LoadNetworkInterfaceList;
end;

procedure TSunriseChatCore.LoadNetworkInterfaceList;
var
  KeyList: TStringList;
  I, II: Integer;
  NetKey: string;
  PnpInstanceID: string;
  Buf: string;
  IPParts: TArrayOfString;
  SubnetParts: TArrayOfString;
begin
  SetLength(NetworkInterfaces,0);     // Clear list
  if Win32Platform > 1 then begin
  KeyList:= TStringList.Create;
  with TRegistry.Create do try
    RootKey := HKEY_LOCAL_MACHINE;
    OpenKey('\SYSTEM\CurrentControlSet\Control\Network',False);
    GetKeyNames(KeyList);
    for I:= 0 to KeyList.Count-1 do begin
      OpenKey('\SYSTEM\CurrentControlSet\Control\Network\'+KeyList[I],False);
      if ReadString('Class') = 'Net' then Break;
    end;
    NetKey:= KeyList[I];
    GetKeyNames(KeyList);

    for I:= 0 to KeyList.Count-1 do begin
      OpenKey('\SYSTEM\CurrentControlSet\Control\Network\'+NetKey+'\'+KeyList[I],False);
      if KeyExists('Connection') then begin
        OpenKey('\SYSTEM\CurrentControlSet\Control\Network\'+NetKey+'\'+KeyList[I]+'\Connection',False);
        //if (ValueExists('MediaSubType') and (ReadInteger('MediaSubType') = 1)) or not ValueExists('MediaSubType') then
        begin
          SetLength(NetworkInterfaces,Length(NetworkInterfaces)+1);
          with NetworkInterfaces[High(NetworkInterfaces)] do begin
            GUID:= KeyList[I];

            // Read network interface name
            Name:= ReadString('Name');
            PnpInstanceID:= ReadString('PnpInstanceID');
            with TRegistry.Create do try
              RootKey := HKEY_LOCAL_MACHINE;

              // Read network device name
              //OpenKey('\SYSTEM\CurrentControlSet\Enum\'+PnpInstanceID,False);
              // ShowMessage(BoolToStr(OpenKey('\SYSTEM\CurrentControlSet\Enum',False)));
              //DeviceName:= ReadString('DeviceDesc');

              OpenKey('\SYSTEM\CurrentControlSet\Services\Tcpip\Parameters\Interfaces\'+KeyList[I],False);
              if KeyExists('\SYSTEM\CurrentControlSet\Services\Tcpip\Parameters\Interfaces\'+KeyList[I]) then begin
                // Read IP address
                SetLength(Buf,GetDataSize('IPAddress'));
                ReadBinaryData('IPAddress',Buf[1],GetDataSize('IPAddress'));
                if Pos(#0,Buf)>0 then IPAddress:= Copy(Buf,1,Pos(#0,Buf)-1)
                else IPAddress:= Copy(Buf,1,Length(Buf));

                // Read subnet mask
                SetLength(Buf,GetDataSize('SubnetMask'));
                ReadBinaryData('SubnetMask',Buf[1],GetDataSize('SubnetMask'));
                if Pos(#0,Buf)>0 then SubnetMask:= Copy(Buf,1,Pos(#0,Buf)-1)
                else SubnetMask:= Copy(Buf,1,Length(Buf));

                // Check DHCP values
                if IPAddress = '0.0.0.0' then begin
                  IPAddress:= ReadString('DhcpIPAddress');
                  SubNetMask:= ReadString('DhcpSubNetMask');
                end;
              end;
              if (IPAddress = '0.0.0.0') or (IPAddress = '') then begin
                IPAddress:= '127.0.0.1';
                SubNetMask:= '255.0.0.0';
              end;
              SubnetParts:= Explode('.', SubNetMask);
              IPParts:= Explode('.', IPAddress);
              for II:= 0 to 3 do
                IPParts[II]:= IntToStr(StrToInt(IPParts[II]) or (StrToInt(SubnetParts[II]) xor $ff));
              BroadcastIPAddress:= IPParts[0]+'.'+IPParts[1]+'.'+IPParts[2]+'.'+IPParts[3];
            finally
              Free;
            end;
          end;
        end;
      end;
    end;
  finally
    Free;
  end;
    KeyList.Free;
    SetLength(Buf,0);
  end else begin
    // For Windows 9x compatibility
    SetLength(NetworkInterfaces,Length(LocalIPAddressesStr));
    for I:= 0 to High(LocalIPAddressesStr) do with NetworkInterfaces[I] do begin
      IPAddress:= LocalIPAddressesStr[I];
      Name:= LocalIPAddressesStr[I];
      GUID:= LocalIPAddressesStr[I];
    end;
  end;
end;

procedure TSunriseChatCore.SelectNetworkInterface(SelectedGUID: string);
var
  I: Integer;
begin
  I:= 0;
  while (I<Length(NetworkInterfaces)) and (NetworkInterfaces[I].GUID <> SelectedGUID) do I:= I + 1;
  if I<Length(NetworkInterfaces) then begin
    NetworkInterfaceIndex:= I;
    LocalUser.IP:= NetworkInterfaces[I].IPAddress;
  end else begin
    if Length(NetworkInterfaces)>0 then begin
      NetworkInterfaceIndex:= 0;
      LocalUser.IP:= NetworkInterfaces[0].IPAddress;
    end else begin
      NetworkInterfaceIndex:= 0;
      LocalUser.IP:= '127.0.0.1';
    end;
  end;
  StartNetwork;
end;

function TSunriseChatCore.Explode(Separator: Char;
  Data: string): TArrayOfString;
begin
  SetLength(Result,0);
  while Pos(Separator,Data)>0 do begin
    SetLength(Result, Length(Result)+1);
    Result[High(Result)]:= Copy(Data,1,Pos(Separator,Data)-1);
    Delete(Data,1,Pos(Separator,Data));
  end;
  SetLength(Result, Length(Result)+1);
  Result[High(Result)]:= Data;
end;

procedure TSunriseChatCore.fndUDPClientSocket1Error(Sender: ASocket);
begin
//  ShowMessage('UDPClient error');
end;

procedure TSunriseChatCore.Timer3Timer(Sender: TObject);
begin
  try
  MainWindow.Caption := GetEnumName(TypeInfo(TSocketState), Integer(fndTCPClient1.Socket.State)) +
  ','+GetEnumName(TypeInfo(TSocketState), Integer(fndUDPClientSocket1.State))+ BoolToStr(fndUDPClientSocket1.Bound);
  finally
  end;
end;

end.
